/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: flate.c,v 1.18 2021/03/14 18:26:47 babolo Exp $"

#define BLIN_COMPAT     3
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/recobe.h>
#include <babolo/parser.h>
#include <mife.h>
#include "pngtr.h"

#define ZCMAX   3
#define ZEMAX   7

typedef struct pngtr_place {
    u_int dx;
    u_int dy;
    u_int sx;
    u_int sy;
} pngtr_place;

static const char *zc[ZCMAX] = {"OK", "STREAM_END", "NEED_DICT"};
static const char *ze[ZEMAX] = {NULL, "", "STREAM", "DATA", "MEM", "BUF", "VERSION"};
static char        buf[14];
static pngtr_place places[] =
{ { 1, 1, 0, 0} /* 0 linear */
, { 8, 8, 0, 0} /* 1 pass */
, { 8, 8, 4, 0} /* 2 pass */
, { 4, 8, 0, 4} /* 3 pass */
, { 4, 4, 2, 0} /* 4 pass */
, { 2, 4, 0, 2} /* 5 pass */
, { 2, 2, 1, 0} /* 6 pass */
, { 1, 2, 0, 1} /* 7 pass */
};

static const char *
zerr(int ret) {
    if  ((ZCMAX <= ret) || (ZEMAX <= -ret)) {
        snprintf(buf, 16, "%dZ", ret);
        return(buf);
    }
    if  (0 <= ret) return(zc[ret]);
    if  (!!ze[-ret]) return(ze[-ret]);
    return(strerror(errno));
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    .     .                              ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_infla(pngtr_conf *cf, u_int chctr, u_int opctr) {                                          /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey = &cf->convey[cf->num];
    pngtr_chunk      *chunk  = &convey->chunks[chctr];
    u_int32_t         nowsz;
    z_stream          strm;
    int               ex     = EX_OK;
    pngtr_operand    *op     = &chunk->operands[opctr];

    ifBLIN_QX3("+ %08X q%u c%u o%u", cf->num, cf->flags, chctr, opctr);
    bzero(&strm, sizeof(z_stream));
    nowsz = PNGTR_CHUNKSZ;
    PNGTR_X(!!(ex = inflateInit(&strm)),            ("inflateInit: %s", zerr(ex)), ENXIO, EX_SOFTWARE);
#if ZLIB_VERNUM > 0x12a0
    if  (!!(PNGTR_NOINVAL & convey->flags)) {
        PNGTR_X( (0 > (ex = inflateValidate(&strm, 0)))
               , ("inflateValidate: %s", zerr(ex))
               , ENXIO
               , EX_SOFTWARE
               );
    }
#endif
    free(chunk->infla);
    chunk->infla = NULL;
    chunk->inflen = 0;
    strm.avail_in = op->size * op->maxindex;
    strm.next_in = op->val;
    nowsz = (op->size * 2) | ((PNGTR_CHUNKSZ - 1) + 1 + PNGTR_CHUNKSZ);
    do {
        chunk->infla = realloc(chunk->infla, (nowsz *= 2));
        chunk->flags |= PNGTR_INFLA;
        strm.next_out = (u_char *)(chunk->infla) + chunk->inflen;
        strm.avail_out = nowsz - chunk->inflen;
        ex = inflate(&strm, Z_NO_FLUSH);
        ifBLIN_QX4("inflate %s %u", zerr(ex), strm.avail_out);
        PNGTR_X((0 > ex),                               ("inflate %s", zerr(ex)), EFTYPE, EX_SOFTWARE);
        chunk->inflen = nowsz - strm.avail_out;
    } while (strm.avail_out == 0);
    if  (Z_STREAM_END == ex) {
        ex = EX_OK;
        (void)inflateEnd(&strm);
    } else if (Z_OK == ex) {
        ex = EX_OK;
    } else {
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (chunk->inflen < nowsz) chunk->infla = realloc(chunk->infla, chunk->inflen);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****      .                                 ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_imgfla(pngtr_conf *cf) {                                                                   /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey = &cf->convey[cf->num];
    int               ex     = EX_OK;

    ifBLIN_QX3("+ %08X %08X %u", cf->flags, convey->flags, cf->num);
    if  (!convey->IHDR.depth || !convey->IHDR.width || !convey->IHDR.height) {
        convey->nowsz = PNGTR_CHUNKSZ;
    } else {
        convey->nowsz = ( ( convey->IHDR.width
                          * (pngtr_pointsize[convey->IHDR.colortype][convey->IHDR.depth])
                          + 31
                          )
                        / 8
                        * convey->IHDR.height
                        + ( !convey->IHDR.interlace
                          ? convey->IHDR.height
                          : ((convey->IHDR.height + 7) / 8 * 15)
                        ) )
        ;
    }
    ifBLIN_QX4("%u", convey->nowsz);
loop:{  u_char *ift;

        PNGTR_X((ex = inflateInit(&convey->strm)),  ("inflateInit: %s", zerr(ex)), ENXIO, EX_SOFTWARE);
#if ZLIB_VERNUM > 0x12a0
        if  (!!(PNGTR_NOINVAL & convey->flags)) {
            PNGTR_X( (0 > (ex = inflateValidate(&convey->strm, 0)))
                   , ("inflateValidate: %s", zerr(ex))
                   , ENXIO
                   , EX_SOFTWARE
                   );
        }
#endif
        ift = convey->imda;
        convey->imda = realloc(convey->imda, (convey->nowsz *= 2));
        PNGTR_W(!convey->imda,                                                ("realloc z"), EX_OSERR);
        convey->strm.next_out += convey->imda - ift;
        convey->strm.avail_out = convey->nowsz;
        for (u_int chctr = 0; chctr < convey->chunum; ++chctr) {
            pngtr_chunk      *chunk = &convey->chunks[chctr];

            if  (PNGTR_IDAT == chunk->type) {
                pngtr_operand    *op     = &chunk->operands[0];

                chunk->infla = convey->strm.next_out;
                convey->strm.avail_in = op->size * op->maxindex;
                convey->strm.next_in = op->val;
                ex = inflate(&convey->strm, Z_NO_FLUSH);
                ifBLIN_QX4("inflate %s %u", zerr(ex), convey->strm.avail_out);
                PNGTR_X( (0 > ex)
                       , ("in %s: inflate %s", !convey->png.nm ? "" : convey->png.nm, zerr(ex))
                       , 0
                       , EX_OK
                       );
                chunk->inflen = convey->strm.next_out - chunk->infla;
                if  (!convey->strm.avail_out) {
                    ifBLIN_QX0( "reinflate %u * 2 %s %u"
                              , convey->nowsz
                              , !convey->png.nm ? "" : convey->png.nm
                              , chctr
                              );
                    goto loop;
                }
                if  (Z_STREAM_END == ex) {
                    (void)inflateEnd(&convey->strm);
                    ex = EX_OK;
                } else if (Z_OK != ex) {
                    ifBLIN_QX0("inflate %s %s", zerr(ex), !convey->png.nm ? "" : convey->png.nm);
                    ex = EX_OK;
                    goto out;
            }   }
            if  (PNGTR_IEND == chunk->type) {
                (void)inflateEnd(&convey->strm);
        }   }
        ifBLIN_QX4("%u %u", convey->nowsz, convey->nowsz - convey->strm.avail_out);
        /*     realloc(),        */
        if  (!(convey->nowsz -= convey->strm.avail_out)) {
            free(convey->imda);
            convey->imda = NULL;
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static u_int8_t
/*****************************************************************************************************
                                                                                                  ****/
PaethPredictor(u_int32_t a, u_int32_t b, u_int32_t c) {                                          /****
                                                                                                  ****
 *****************************************************************************************************/
    int p  = (int32_t)a + (int32_t)b - (int32_t)c;
    int pa = abs(p - (int32_t)a);
    int pb = abs(p - (int32_t)b);
    int pc = abs(p - (int32_t)c);

    if  ((pa <= pb) && (pa <= pc)) return(a);
    if  (pb <= pc) return(b);
    return(c);
}

static int
/*****************************************************************************************************
                                                                                                  ****/
defilt(pngtr_conf *cf, u_int ins, int ilace, FILE *pr) {                                         /****
                                                                                                  ****
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey = &cf->convey[cf->num];
    u_int             width;                                 /*                */
    u_int             nowpx;                                 /*         */
    u_int             noww;                                  /*           */
    u_int             nowh;                                  /*          */
    u_int             step;                                  /*              */
    u_int             bpp;                                   /*          */
    u_int             row;                                   /*      */
    u_int8_t         *out;
    int               ex = 0;
    u_int             dx = places[ilace].dx;
    u_int             dy = places[ilace].dy;
    u_int             sx = places[ilace].sx;
    u_int             sy = places[ilace].sy;

    bpp = pngtr_pointsize[convey->IHDR.colortype][convey->IHDR.depth];
    step = (bpp + 7) / 8;
    nowpx = ((convey->IHDR.width + dx - sx - 1) / dx);
    noww = (nowpx * bpp + 7) / 8;
    nowh = (convey->IHDR.height + dy - sy - 1) / dy;
    ifBLIN_QX3( "+ %08X c%u s%u b%u %u(%u)x%u (%ux%u %ux%u)"
              , cf->flags
              , cf->num
              , step
              , bpp
              , nowpx
              , noww
              , nowh
              , dx
              , dy
              , sx
              , sy
              );
    if  (!noww || !nowh) goto out;
    if  (8 > bpp) {
        width = (convey->IHDR.width * bpp + 7) / 8;
        PNGTR_W( !(out = calloc(convey->IHDR.height, width))
               , ("calloc %u x %u", convey->IHDR.height, width)
               , EX_OSERR
               );
#       define I(R, C) convey->imda[ins + ((noww + 1) * (R)) + (C) + 1]
#       define O(R, C) out[noww * (R) + (C)]
        for (u_int r = 0; r < nowh; ++r) {
            row = (noww + 1) * r + ins;
            if  (!!pr) fprintf(pr, "%u", convey->imda[row]);
            ex = row + noww + 1;
            ifBLIN_QX4("%04X %04X", row, ex);
            for (u_int c = 0; c < noww; ++c) {
                switch(convey->imda[row]) {
                case 0:
                    O(r, c) = I(r, c);
                    break;
                case 1:
                    O(r, c) = I(r, c) + ((c < 1) ? 0 : O(r, c - 1));
                    break;
                case 2:
                    O(r, c) = I(r, c) + (!r ? 0 : O(r - 1, c));
                    break;
                case 3:
                    O(r, c) = I(r, c) + ((!r ? 0 : O(r - 1, c)) + ((c < 1) ? 0 : O(r, c - 1))) / 2;
                    break;
                case 4:
                    O(r, c) = I(r, c) + PaethPredictor( (c < 1) ? 0 : O(r, c - 1)
                                                      , !r ? 0 : O(r - 1, c)
                                                      , ((c < 1) || !r) ? 0 : O(r - 1, c - 1)
                                                      )
                    ;
                    break;
                default:
                    ifBLIN_QX0("Illegal filter %u @%04X", convey->imda[row], row);
                    ex = -EX_USAGE;
                    errno = EINVAL;
                    goto out;
        }   }   }
#       undef O
#       undef I
        if  (0x80 & (convey->flags << ilace)) goto out;            /*    */
        switch(bpp) {
        case 1:
            for (u_int r = 0; r < nowh; ++r) {
                for (u_int c = 0; c < nowpx; ++c) {
                    if  (out[noww * r + (c / 8)] & (0x80 >> (c & 7))) {
                           convey->img.cont[width * (r * dy + sy) + (c * dx + sx) / 8]
                        |= (0x80 >> ((c * dx + sx) & 7));
            }   }   }
            break;
        case 2:
            for (u_int r = 0; r < nowh; ++r) {
                for (u_int c = 0; c < nowpx; ++c) {
                    ifBLIN_QX4( "%ux%u %u %02X %02X %u %02X %02X"
                              , r
                              , c
                              , noww * r + (c / 4)
                              , out[noww * r + (c / 4)]
                              , 0xC0 >> ((c & 3) * 2)
                              , 6 - ((c * dx + sx) & 3) * 2
                              ,    (out[noww * r + (c / 4)] & (0xC0 >> ((c & 3) * 2)))
                                >> (6 - ((c & 3) * 2))
                              , (  (  (out[noww * r + (c / 4)] & (0xC0 >> ((c & 3) * 2)))
                                   >> (6 - ((c & 3) * 2))
                                   )
                                << (6 - ((c * dx + sx) & 3) * 2)
                                ) & 0xFF
                              );
                       convey->img.cont[width * (r * dy + sy) + (c * dx + sx) / 4]
                    |=    ((out[noww * r + (c / 4)] & (0xC0 >> ((c & 3) * 2))) >> (6 - ((c & 3) * 2)))
                       << (6 - ((c * dx + sx) & 3) * 2)
                    ;
            }   }
            break;
        case 4:
            for (u_int r = 0; r < nowh; ++r) {
                for (u_int c = 0; c < nowpx; ++c) {
                       convey->img.cont[width * (r * dy + sy) + (c * dx + sx) / 2]
                    |=    (out[noww * r + (c / 2)] & (0xF0 >> ((c & 1) * 4))) >> (4 - ((c & 1) * 4))
                       << (4 - ((c * dx + sx) & 1) * 4)
                    ;
            }   }
            break;
        }
        free(out);
    } else {
#       define I(R, C, S) convey->imda[(noww + 1) * (R) + ((C) * step) + ins + (S) + 1]
#       define O(R, C, S) out[(convey->IHDR.width * ((R) * dy + sy) + sx + dx * (C)) * step + (S)]
        out = convey->img.cont;
        for (u_int r = 0; r < nowh; ++r) {
            row = (noww + 1) * r + ins;
            if  (!!pr) fprintf(pr, "%u", convey->imda[row]);
            ex = row + noww + 1;
            ifBLIN_QX4("%04X %04X", row, ex);
            for (u_int c = 0; c < convey->IHDR.width; ++c) {
                switch(convey->imda[row]) {
                case 0:
                    for (u_int s = 0; s < step; ++s) {
                        O(r, c, s) = I(r, c, s);
                    }
                    break;
                case 1:
                    for (u_int s = 0; s < step; ++s) {
                        O(r, c, s) = I(r, c, s) + ((c < 1) ? 0 : O(r, c - 1, s));
                    }
                    break;
                case 2:
                    for (u_int s = 0; s < step; ++s) {
                        O(r, c, s) = I(r, c, s) + (!r ? 0 : O(r - 1, c, s));
                    }
                    break;
                case 3:
                    for (u_int s = 0; s < step; ++s) {
                        O(r, c, s) = I(r, c, s)
                                   + ((!r ? 0 : O(r - 1, c, s)) + ((c < 1) ? 0 : O(r, c - 1, s)))
                                   / 2
                        ;
                    }
                    break;
                case 4:
                    for (u_int s = 0; s < step; ++s) {
                        O(r, c, s) = I(r, c, s)
                                   + PaethPredictor( (c < 1) ? 0 : O(r, c - 1, s)
                                                   , !r ? 0 : O(r - 1, c, s)
                                                   , ((c < 1) || !r) ? 0 : O(r - 1, c - 1, s)
                                                   )
                        ;
                    }
                    break;
                default:
                    ifBLIN_QX0("Illegal filter %u @%04X", convey->imda[row], row);
                    ex = -EX_USAGE;
                    errno = EINVAL;
                    goto out;
        }   }   }
#       undef O
#       undef I
    }

out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****  .                                                                   ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_flimg(pngtr_conf *cf, FILE *pr) {                                                          /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey = &cf->convey[cf->num];
    u_int32_t         bpp    = pngtr_pointsize[convey->IHDR.colortype][convey->IHDR.depth];
    u_int32_t         width  = (convey->IHDR.width * bpp + 7) / 8;
    u_int             ins;
    int               ex     = EX_OK;

    ifBLIN_QX3("+ %08X %u", cf->flags, cf->num);
    if  (!!convey->img.cont) goto out;
    PNGTR_W( !(convey->img.cont = calloc(convey->IHDR.height, width))
           , ("calloc %u x %u", convey->IHDR.height, width)
           , EX_OSERR
           );
    convey->img.size = convey->IHDR.height * width;
    switch(convey->IHDR.interlace) {
    case 0:
        PNGTR_O( 0 > (ex = defilt(cf, 0, 0, pr))
               , ("defilt %s", !convey->png.nm ? "" : convey->png.nm)
               );
        ins = ex;
        ex = 0;
        break;
    case 1:
        ins = 0;
        for (u_int p = 1; p <= 7; ++p) {
            ex = defilt(cf, ins, p, pr);
            PNGTR_O((0 > ex),                    ("defilt %s", !convey->png.nm ? "" : convey->png.nm));
            ins = ex;
        }
        ex = 0;
        break;
    default:
            ifBLIN_QX0("Illegal interlace %u", convey->IHDR.interlace);
            ex = EX_USAGE;
            errno = EINVAL;
            goto out;
    }
    if  (convey->nowsz != ins) {
        ifBLIN_QX0( "End of work %06X <> end of in %06X in %s"
                  , BLIN_I(ins)
                  , BLIN_I(convey->nowsz)
                  , !convey->png.nm ? "" : convey->png.nm
                  );
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     num-                                            ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_realimg(pngtr_conf *cf) {                                                                  /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    u_int8_t         *defimg;
    pngtr_convey     *convey;
    pngtr_chunk      *chunk;
    u_int32_t         outsz;
    u_int32_t         width;
    z_stream          strm;
    u_int32_t         bpp;
    pngtr_operand    *op;
    int               ex    = EX_OK;

    ifBLIN_QX3("+ %08X %u", cf->flags, cf->num);
    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    convey = &cf->convey[cf->num];
    bpp    = pngtr_pointsize[convey->IHDR.colortype][convey->IHDR.depth];
    width  = (convey->IHDR.width * bpp + 7) / 8;
    convey->nowsz = (width + 1) * convey->IHDR.height;
    PNGTR_W(!(convey->imda = malloc(convey->nowsz)),           ("malloc %u", convey->nowsz), EX_OSERR);
    for (u_int i = 0; i < convey->IHDR.height; ++i) {
        convey->imda[(width + 1) * i] = 0;
        bcopy(&convey->img.cont[width * i], &convey->imda[(width + 1) * i + 1], width);
    }
    outsz = convey->nowsz + 12;
    PNGTR_W(!(defimg = malloc(outsz)),                                 ("malloc %u", outsz), EX_OSERR);
    bzero(&strm, sizeof(z_stream));
    PNGTR_X(!!(ex = deflateInit(&strm, 9/*lvl*/)),  ("deflateInit: %s", zerr(ex)), ENXIO, EX_SOFTWARE);
    strm.avail_in = convey->nowsz;
    strm.next_in = convey->imda;
    strm.avail_out = outsz;
    strm.next_out = defimg;
    ex = deflate(&strm, Z_FINISH);
    ifBLIN_QX4("deflate %s %u", zerr(ex), strm.avail_out);
    PNGTR_X((0 > ex),                                   ("deflate %s", zerr(ex)), EFTYPE, EX_SOFTWARE);
    if  (Z_STREAM_END == ex) {
        ex = EX_OK;
        (void)deflateEnd(&strm);
    } else if (Z_OK == ex) {
        ex = EX_OK;
    } else {
        ex = EX_SOFTWARE;
        goto out;
    }
    outsz -= strm.avail_out;
    for (u_int chctr = 0; chctr <= convey->chunum; ++chctr) {
        if  (!convey->chunks) continue;
        if  (!(chunk = &convey->chunks[chctr])) continue;
        if  (PNGTR_IDAT != chunk->type) continue;
        break;
    }
    if  (!chunk->operands) {
        PNGTR_W(!(chunk->operands = calloc(2, sizeof(pngtr_operand))), ("calloc 2 operand"), EX_OSERR);
    }
    op = &chunk->operands[0];
    if  (PNGTR_VIRTUAL & op->flags) free(op->val);
    PNGTR_W(!(op->val = realloc(defimg, outsz)),                      ("realloc %u", outsz), EX_OSERR);
    op->maxindex = outsz;
    op->size = 1;
    op->flags |= PNGTR_VIRTUAL;
    chunk->flags |= PNGTR_VIRTUAL;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
